#include <stdio.h>
#include <stdarg.h>

#include <qdir.h>

#if !defined(NDEBUG)

#include "TraceManager.h"

#include "TraceModule.h"


/**
  The tracemodule is the object that gets instantiated in your source file.
  It forwards log events to the TraceManager
*/

using namespace NTracer;

/**
  \brief Constructor; there is no default constructor.

  \param mgr Pointer to the manager object
  \param name The name of the module
  \param level Initial level
*/
CTraceModule::CTraceModule(CTraceManager *mgr, const QString &name, int level)
{
   m_pManager = mgr;
   ModuleName = name;
   m_Level = level;
}


QString CTraceModule::GetName() const
{
   return ModuleName;
}

int CTraceModule::GetLevel() const
{
   return m_Level;
}


/**
  \brief Log function enter point

  This function marks the entry of a function. It prefixes the function name with
  >> and increments the indent level.

*/
void CTraceModule::Enter(const QString &function_name)
{
   m_pManager->Log(">> " + function_name);
   m_pManager->IncIndent();
}

/**
  \brief Log function enter with the arguments

  This function allows you to create log entries of the form:
  \code
  >> MyFunction(a = 45, b = 3, string = 'hello')
  \endcode

  Of course you don't have to specify just the arguments supplied to the function,
  but it's good practice.

  This function is normally called by using the TR_FUNC(arg, arg...) macro
*/
void CTraceModule::Enter(const QString &function_name, const QString &args)
{
  m_pManager->Log(">> " + function_name + "(" + args + ")");
  m_pManager->IncIndent();
}

void CTraceModule::Enter(const QString &file_name, int line_number)
{
   int sep;
   //Enter(file_name.section(QDir::separator(), -1) + ":" + QString::number(line_number));
   sep = file_name.findRev(QDir::separator());
   Enter(file_name.mid(sep + 1) + ":" + QString::number(line_number));
}

/**
  \brief Mark function exit point

  Calling this function marks the end of the current function. It decrements the logging
  level and logs a line of the form:

  \code
  << MyFunction()
  \endcode

*/
void CTraceModule::Leave(const QString &function_name)
{
   m_pManager->DecIndent();
   m_pManager->Log("<< " + function_name);
}

void CTraceModule::Leave(const QString &file_name, int line_number)
{
   int sep;
   //Leave(file_name.section(QDir::separator(), - 1) + ":" + QString::number(line_number));
   sep = file_name.findRev(QDir::separator());
   Leave(file_name.mid(sep + 1) + ":" + QString::number(line_number));
}

void CTraceModule::Log(int level, const char *fmt, ...)
{
   va_list args;

   if (level > m_Level)
     return;

   /* Unfortunately we cannot propagate the ellipses (...), so we must build
      the string here */
   va_start(args, fmt);
#if defined(_WIN32)
   _vsnprintf(LargeBuffer, 2048, fmt, args);
#else
   vsnprintf(LargeBuffer, 2048, fmt, args);
#endif
   va_end(args);
   m_pManager->Log(LargeBuffer);
}


void CTraceModule::IncLevel()
{
   m_Level++;
}

void CTraceModule::DecLevel()
{
   if (m_Level > 0)
     m_Level--;
}

void CTraceModule::SetLevel(int level)
{
   if (level >= 0)
     m_Level = level;
}

#endif
